@@ -9,7 +9,23 @@ all rights reserved.
 let poly_compare=compare
 open Printf
 
-module RS = Random.State
+module RS = struct
+  (* Poor man's splitter for version < 5.0                       *)
+  (* This definition is shadowed by the [include] on OCaml >=5.0 *)
+  let split rs =
+    let bits = Random.State.bits rs in
+    let rs' = Random.State.make [|bits|] in
+    rs'
+  include Random.State
+  (* This is how OCaml 5.0 splits:             *)
+  (* Split a new PRNG off the given PRNG *)
+  (*
+  let split s =
+    let i1 = bits64 s in let i2 = bits64 s in
+    let i3 = bits64 s in let i4 = bits64 s in
+    mk i1 i2 i3 i4
+  *)
+end
 
 let (|>) x f = f x
 
@@ -42,12 +58,43 @@ let _opt_map_4 ~f a b c d = match a, b, c, d with
   | Some x, Some y, Some z, Some w -> Some (f x y z w)
   | _ -> None
 
+let _opt_map_5 ~f a b c d e = match a, b, c, d, e with
+  | Some x, Some y, Some z, Some u, Some v -> Some (f x y z u v)
+  | _ -> None
+
+let _opt_map_6 ~f a b c d e g = match a, b, c, d, e, g with
+  | Some a, Some b, Some c, Some d, Some e, Some g -> Some (f a b c d e g)
+  | _ -> None
+
+let _opt_map_7 ~f a b c d e g h = match a, b, c, d, e, g, h with
+  | Some a, Some b, Some c, Some d, Some e, Some g, Some h -> Some (f a b c d e g h)
+  | _ -> None
+
+let _opt_map_8 ~f a b c d e g h i = match a, b, c, d, e, g, h, i with
+  | Some a, Some b, Some c, Some d, Some e, Some g, Some h, Some i ->
+    Some (f a b c d e g h i)
+  | _ -> None
+
+let _opt_map_9 ~f a b c d e g h i j = match a, b, c, d, e, g, h, i, j with
+  | Some a, Some b, Some c, Some d, Some e, Some g, Some h, Some i, Some j ->
+    Some (f a b c d e g h i j)
+  | _ -> None
+
 let _opt_sum a b = match a, b with
   | Some _, _ -> a
   | None, _ -> b
 
 let sum_int = List.fold_left (+) 0
 
+(* Included for backwards compatibility, pre 4.13 *)
+let string_fold_right f s acc =
+  let len = String.length s in
+  let rec loop i acc =
+    if i<0
+    then acc
+    else loop (i-1) (f s.[i] acc) in
+  loop (len-1) acc
+
 exception No_example_found of string
 (* raised if an example failed to be found *)
 
@@ -136,11 +183,13 @@ module Gen = struct
 
   let neg_int st = -(nat st)
 
-  let opt ?(ratio = 0.85) f st =
+  let option ?(ratio = 0.85) f st =
     let p = RS.float st 1. in
     if p < (1.0 -. ratio) then None
     else Some (f st)
 
+  let opt = option
+
   (* Uniform random int generator *)
   let pint =
     if Sys.word_size = 32 then
@@ -283,6 +332,27 @@ module Gen = struct
 
   let char st = char_of_int (RS.int st 256)
 
+  let tup2 = pair
+
+  let tup3 = triple
+
+  let tup4 = quad
+
+  let tup5 (g1 : 'a t) (g2 : 'b t) (g3 : 'c t) (g4 : 'd t) (g5 : 'e t) : ('a * 'b * 'c * 'd * 'e) t =
+    (fun a b c d e -> (a, b, c, d, e)) <$> g1 <*> g2 <*> g3 <*> g4 <*> g5
+
+  let tup6 (g1 : 'a t) (g2 : 'b t) (g3 : 'c t) (g4 : 'd t) (g5 : 'e t) (g6 : 'f t) : ('a * 'b * 'c * 'd * 'e * 'f) t =
+    (fun a b c d e f -> (a, b, c, d, e, f)) <$> g1 <*> g2 <*> g3 <*> g4 <*> g5 <*> g6
+
+  let tup7 (g1 : 'a t) (g2 : 'b t) (g3 : 'c t) (g4 : 'd t) (g5 : 'e t) (g6 : 'f t) (g7 : 'g t) : ('a * 'b * 'c * 'd * 'e * 'f * 'g) t =
+    (fun a b c d e f g -> (a, b, c, d, e, f, g)) <$> g1 <*> g2 <*> g3 <*> g4 <*> g5 <*> g6 <*> g7
+
+  let tup8 (g1 : 'a t) (g2 : 'b t) (g3 : 'c t) (g4 : 'd t) (g5 : 'e t) (g6 : 'f t) (g7 : 'g t) (g8 : 'h t) : ('a * 'b * 'c * 'd * 'e * 'f * 'g * 'h) t =
+    (fun a b c d e f g h -> (a, b, c, d, e, f, g, h)) <$> g1 <*> g2 <*> g3 <*> g4 <*> g5 <*> g6 <*> g7 <*> g8
+
+  let tup9 (g1 : 'a t) (g2 : 'b t) (g3 : 'c t) (g4 : 'd t) (g5 : 'e t) (g6 : 'f t) (g7 : 'g t) (g8 : 'h t) (g9 : 'i t) : ('a * 'b * 'c * 'd * 'e * 'f * 'g * 'h * 'i) t =
+    (fun a b c d e f g h i -> (a, b, c, d, e, f, g, h, i)) <$> g1 <*> g2 <*> g3 <*> g4 <*> g5 <*> g6 <*> g7 <*> g8 <*> g9
+
   let printable_chars =
     let l = 126-32+1 in
     let s = Bytes.create l in
@@ -295,19 +365,31 @@ module Gen = struct
   let printable st = printable_chars.[RS.int st (String.length printable_chars)]
   let numeral st = char_of_int (48 + RS.int st 10)
 
-  let string_size ?(gen = char) size st =
+  let bytes_size ?(gen = char) size st =
     let s = Bytes.create (size st) in
     for i = 0 to Bytes.length s - 1 do
       Bytes.set s i (gen st)
     done;
+    s
+
+  let string_size ?(gen = char) size st =
+    let s = bytes_size ~gen size st in
     Bytes.unsafe_to_string s
+
+  let bytes ?gen st = bytes_size ?gen nat st
   let string ?gen st = string_size ?gen nat st
+  let bytes_of gen = bytes_size ~gen nat
   let string_of gen = string_size ~gen nat
+  let bytes_printable = bytes_size ~gen:printable nat
   let string_printable = string_size ~gen:printable nat
   let string_readable = string_printable
+  let bytes_small st = bytes_size small_nat st
+  let bytes_small_of gen st = bytes_size ~gen small_nat st
   let small_string ?gen st = string_size ?gen small_nat st
   let small_list gen = list_size small_nat gen
   let small_array gen = array_size small_nat gen
+  let string_small st = string_size small_nat st
+  let string_small_of gen st = string_size ~gen small_nat st
 
   let join g st = (g st) st
 
@@ -392,6 +474,7 @@ module Print = struct
   let int = string_of_int
   let bool = string_of_bool
   let float = string_of_float
+  let bytes = Bytes.to_string
   let string s = s
   let char c = String.make 1 c
 
@@ -404,6 +487,121 @@ module Print = struct
   let quad a b c d (x,y,z,w) =
     Printf.sprintf "(%s, %s, %s, %s)" (a x) (b y) (c z) (d w)
 
+  let default = fun _ -> "<no printer>"
+
+  let tup2 p_a p_b (a, b) =
+    Printf.sprintf "(%s, %s)" (p_a a) (p_b b)
+
+  let tup2_opt p_a p_b (a, b) =
+    let p_a = Option.value ~default p_a in
+    let p_b = Option.value ~default p_b in
+    tup2 p_a p_b (a, b)
+
+  let tup3 p_a p_b (p_c) (a, b, c) =
+    Printf.sprintf "(%s, %s, %s)" (p_a a) (p_b b) (p_c c)
+
+  let tup3_opt p_a p_b p_c (a, b, c) =
+    let p_a = Option.value ~default p_a in
+    let p_b = Option.value ~default p_b in
+    let p_c = Option.value ~default p_c in
+    tup3 p_a p_b p_c (a, b, c)
+
+  let tup4 p_a p_b p_c p_d (a, b, c, d) =
+    Printf.sprintf "(%s, %s, %s, %s)"
+      (p_a a) (p_b b)
+      (p_c c) (p_d d)
+
+  let tup4_opt p_a p_b p_c p_d (a, b, c, d) =
+    let p_a = Option.value ~default p_a in
+    let p_b = Option.value ~default p_b in
+    let p_c = Option.value ~default p_c in
+    let p_d = Option.value ~default p_d in
+    tup4 p_a p_b p_c p_d (a, b, c, d)
+
+  let tup5 p_a p_b p_c p_d p_e (a, b, c, d, e) =
+    Printf.sprintf "(%s, %s, %s, %s, %s)"
+      (p_a a) (p_b b)
+      (p_c c) (p_d d)
+      (p_e e)
+
+  let tup5_opt p_a p_b p_c p_d p_e (a, b, c, d, e) =
+    let p_a = Option.value ~default p_a in
+    let p_b = Option.value ~default p_b in
+    let p_c = Option.value ~default p_c in
+    let p_d = Option.value ~default p_d in
+    let p_e = Option.value ~default p_e in
+    tup5 p_a p_b p_c p_d p_e (a, b, c, d, e)
+
+  let tup6 p_a p_b p_c p_d p_e p_f (a, b, c, d, e, f) =
+    Printf.sprintf "(%s, %s, %s, %s, %s, %s)"
+      (p_a a) (p_b b)
+      (p_c c) (p_d d)
+      (p_e e) (p_f f)
+
+  let tup6_opt p_a p_b p_c p_d p_e p_f (a, b, c, d, e, f) =
+    let p_a = Option.value ~default p_a in
+    let p_b = Option.value ~default p_b in
+    let p_c = Option.value ~default p_c in
+    let p_d = Option.value ~default p_d in
+    let p_e = Option.value ~default p_e in
+    let p_f = Option.value ~default p_f in
+    tup6 p_a p_b p_c p_d p_e p_f (a, b, c, d, e, f)
+
+  let tup7 p_a p_b p_c p_d p_e p_f p_g (a, b, c, d, e, f, g) =
+    Printf.sprintf "(%s, %s, %s, %s, %s, %s, %s)"
+      (p_a a) (p_b b)
+      (p_c c) (p_d d)
+      (p_e e) (p_f f)
+      (p_g g)
+
+  let tup7_opt p_a p_b p_c p_d p_e p_f p_g (a, b, c, d, e, f, g) =
+    let p_a = Option.value ~default p_a in
+    let p_b = Option.value ~default p_b in
+    let p_c = Option.value ~default p_c in
+    let p_d = Option.value ~default p_d in
+    let p_e = Option.value ~default p_e in
+    let p_f = Option.value ~default p_f in
+    let p_g = Option.value ~default p_g in
+    tup7 p_a p_b p_c p_d p_e p_f p_g (a, b, c, d, e, f, g)
+
+  let tup8 p_a p_b p_c p_d p_e p_f p_g p_h (a, b, c, d, e, f, g, h) =
+    Printf.sprintf "(%s, %s, %s, %s, %s, %s, %s, %s)"
+      (p_a a) (p_b b)
+      (p_c c) (p_d d)
+      (p_e e) (p_f f)
+      (p_g g) (p_h h)
+
+  let tup8_opt p_a p_b p_c p_d p_e p_f p_g p_h (a, b, c, d, e, f, g, h) =
+    let p_a = Option.value ~default p_a in
+    let p_b = Option.value ~default p_b in
+    let p_c = Option.value ~default p_c in
+    let p_d = Option.value ~default p_d in
+    let p_e = Option.value ~default p_e in
+    let p_f = Option.value ~default p_f in
+    let p_g = Option.value ~default p_g in
+    let p_h = Option.value ~default p_h in
+    tup8 p_a p_b p_c p_d p_e p_f p_g p_h (a, b, c, d, e, f, g, h)
+
+  let tup9 p_a p_b p_c p_d p_e p_f p_g p_h p_i (a, b, c, d, e, f, g, h, i) =
+    Printf.sprintf "(%s, %s, %s, %s, %s, %s, %s, %s, %s)"
+      (p_a a) (p_b b)
+      (p_c c) (p_d d)
+      (p_e e) (p_f f)
+      (p_g g) (p_h h)
+      (p_i i)
+
+  let tup9_opt p_a p_b p_c p_d p_e p_f p_g p_h p_i (a, b, c, d, e, f, g, h, i) =
+    let p_a = Option.value ~default p_a in
+    let p_b = Option.value ~default p_b in
+    let p_c = Option.value ~default p_c in
+    let p_d = Option.value ~default p_d in
+    let p_e = Option.value ~default p_e in
+    let p_f = Option.value ~default p_f in
+    let p_g = Option.value ~default p_g in
+    let p_h = Option.value ~default p_h in
+    let p_i = Option.value ~default p_i in
+    tup9 p_a p_b p_c p_d p_e p_f p_g p_h p_i (a, b, c, d, e, f, g, h, i)
+
   let list pp l =
     let b = Buffer.create 25 in
     Buffer.add_char b '[';
@@ -511,21 +709,24 @@ module Shrink = struct
 
   let filter f shrink x = Iter.filter f (shrink x)
 
-  let char c yield =
-    if Char.code c > 0 then yield (Char.chr (Char.code c-1))
+  let char_generic target c =
+    if c = target
+    then Iter.empty
+    else
+      let c_code = Char.code c in
+      let target_code = Char.code target in
+      Iter.map (fun diff -> Char.chr (target_code + diff)) (int (c_code - target_code))
+  let char = char_generic 'a'
+  let char_numeral = char_generic '0'
+
+  let char_printable = function
+    | '\n' -> char '~' (* treat '\n' (10) as '~' (126) to ensure a non-trivial, printable output *)
+    | c -> char c
 
   let option s x = match x with
     | None -> Iter.empty
     | Some x -> Iter.(return None <+> map (fun y->Some y) (s x))
 
-  let string s yield =
-    for i =0 to String.length s-1 do
-      let s' = Bytes.init (String.length s-1)
-        (fun j -> if j<i then s.[j] else s.[j+1])
-      in
-      yield (Bytes.unsafe_to_string s')
-    done
-
   let array ?shrink a yield =
     let n = Array.length a in
     let chunk_size = ref n in
@@ -551,37 +752,20 @@ module Shrink = struct
           )
         done
 
-  let list_spine l yield =
-    let n = List.length l in
-    let chunk_size = ref ((n+1)/2) in
-
-    (* push the [n] first elements of [l] into [q], return the rest of the list *)
-    let rec fill_queue n l q = match n,l with
-      | 0, _ -> l
-      | _, x::xs ->
-        Queue.push x q;
-        fill_queue (n-1) xs q
-      | _, _ -> assert false
-    in
-
-    (* remove elements from the list, by chunks of size [chunk_size] (bigger
-       chunks first) *)
-    while !chunk_size > 0 do
-      let q = Queue.create () in
-      let l' = fill_queue !chunk_size l q in
-      (* remove [chunk_size] elements in queue *)
-      let rec pos_loop rev_prefix suffix =
-        yield (List.rev_append rev_prefix suffix);
-        match suffix with
-        | [] -> ()
-        | x::xs ->
-          Queue.push x q;
-          let y = Queue.pop q in
-          (pos_loop [@tailcall]) (y::rev_prefix) xs
-      in
-      pos_loop [] l';
-      chunk_size := !chunk_size / 2;
-    done
+  let rec list_spine l yield =
+    let rec split l len acc = match len,l with
+      | _,[]
+      | 0,_ -> List.rev acc, l
+      | _,x::xs -> split xs (len-1) (x::acc) in
+    match l with
+    | [] -> ()
+    | [_] -> yield []
+    | [x;y] -> yield []; yield [x]; yield [y]
+    | _::_ ->
+      let len = List.length l in
+      let xs,ys = split l ((1 + len) / 2) [] in
+      yield xs;
+      list_spine xs (fun xs' -> yield (xs'@ys))
 
   let list_elems shrink l yield =
     (* try to shrink each element of the list *)
@@ -599,6 +783,18 @@ module Shrink = struct
     | None -> ()
     | Some shrink -> list_elems shrink l yield
 
+  let string ?(shrink = char) s yield =
+    let buf = Buffer.create 42 in
+    list ~shrink
+      (string_fold_right (fun c acc -> c::acc) s [])
+      (fun cs ->
+         List.iter (fun c -> Buffer.add_char buf c) cs;
+         let s = Buffer.contents buf in
+         Buffer.clear buf;
+         yield s)
+
+  let bytes ?(shrink = char) b = Iter.map Bytes.of_string (string ~shrink (Bytes.to_string b))
+
   let pair a b (x,y) yield =
     a x (fun x' -> yield (x',y));
     b y (fun y' -> yield (x,y'))
@@ -613,6 +809,127 @@ module Shrink = struct
     b y (fun y' -> yield (x,y',z,w));
     c z (fun z' -> yield (x,y,z',w));
     d w (fun w' -> yield (x,y,z,w'))
+
+  let default = nil
+
+  let tup2 = pair
+
+  let tup2_opt a b =
+    let a = Option.value ~default a in
+    let b = Option.value ~default b in
+    tup2 a b
+
+  let tup3 = triple
+
+  let tup3_opt a b c =
+    let a = Option.value ~default a in
+    let b = Option.value ~default b in
+    let c = Option.value ~default c in
+    tup3 a b c
+
+  let tup4 = quad
+
+  let tup4_opt a b c d =
+    let a = Option.value ~default a in
+    let b = Option.value ~default b in
+    let c = Option.value ~default c in
+    let d = Option.value ~default d in
+    tup4 a b c d
+
+  let tup5 a b c d e (a', b', c', d', e') yield =
+    a a' (fun x -> yield (x,b',c',d',e'));
+    b b' (fun x -> yield (a',x,c',d',e'));
+    c c' (fun x -> yield (a',b',x,d',e'));
+    d d' (fun x -> yield (a',b',c',x,e'));
+    e e' (fun x -> yield (a',b',c',d',x))
+
+  let tup5_opt a b c d e =
+    let a = Option.value ~default a in
+    let b = Option.value ~default b in
+    let c = Option.value ~default c in
+    let d = Option.value ~default d in
+    let e = Option.value ~default e in
+    tup5 a b c d e
+
+  let tup6 a b c d e f (a', b', c', d', e', f') yield =
+    a a' (fun x -> yield (x,b',c',d',e',f'));
+    b b' (fun x -> yield (a',x,c',d',e',f'));
+    c c' (fun x -> yield (a',b',x,d',e',f'));
+    d d' (fun x -> yield (a',b',c',x,e',f'));
+    e e' (fun x -> yield (a',b',c',d',x,f'));
+    f f' (fun x -> yield (a',b',c',d',e',x))
+
+  let tup6_opt a b c d e f =
+    let a = Option.value ~default a in
+    let b = Option.value ~default b in
+    let c = Option.value ~default c in
+    let d = Option.value ~default d in
+    let e = Option.value ~default e in
+    let f = Option.value ~default f in
+    tup6 a b c d e f
+
+  let tup7 a b c d e f g (a', b', c', d', e', f', g') yield =
+    a a' (fun x -> yield (x,b',c',d',e',f',g'));
+    b b' (fun x -> yield (a',x,c',d',e',f',g'));
+    c c' (fun x -> yield (a',b',x,d',e',f',g'));
+    d d' (fun x -> yield (a',b',c',x,e',f',g'));
+    e e' (fun x -> yield (a',b',c',d',x,f',g'));
+    f f' (fun x -> yield (a',b',c',d',e',x,g'));
+    g g' (fun x -> yield (a',b',c',d',e',f',x))
+
+  let tup7_opt a b c d e f g =
+    let a = Option.value ~default a in
+    let b = Option.value ~default b in
+    let c = Option.value ~default c in
+    let d = Option.value ~default d in
+    let e = Option.value ~default e in
+    let f = Option.value ~default f in
+    let g = Option.value ~default g in
+    tup7 a b c d e f g
+
+  let tup8 a b c d e f g h (a', b', c', d', e', f', g', h') yield =
+    a a' (fun x -> yield (x,b',c',d',e',f',g',h'));
+    b b' (fun x -> yield (a',x,c',d',e',f',g',h'));
+    c c' (fun x -> yield (a',b',x,d',e',f',g',h'));
+    d d' (fun x -> yield (a',b',c',x,e',f',g',h'));
+    e e' (fun x -> yield (a',b',c',d',x,f',g',h'));
+    f f' (fun x -> yield (a',b',c',d',e',x,g',h'));
+    g g' (fun x -> yield (a',b',c',d',e',f',x,h'));
+    h h' (fun x -> yield (a',b',c',d',e',f',g',x))
+
+  let tup8_opt a b c d e f g h =
+    let a = Option.value ~default a in
+    let b = Option.value ~default b in
+    let c = Option.value ~default c in
+    let d = Option.value ~default d in
+    let e = Option.value ~default e in
+    let f = Option.value ~default f in
+    let g = Option.value ~default g in
+    let h = Option.value ~default h in
+    tup8 a b c d e f g h
+
+  let tup9 a b c d e f g h i (a', b', c', d', e', f', g', h', i') yield =
+    a a' (fun x -> yield (x,b',c',d',e',f',g',h',i'));
+    b b' (fun x -> yield (a',x,c',d',e',f',g',h',i'));
+    c c' (fun x -> yield (a',b',x,d',e',f',g',h',i'));
+    d d' (fun x -> yield (a',b',c',x,e',f',g',h',i'));
+    e e' (fun x -> yield (a',b',c',d',x,f',g',h',i'));
+    f f' (fun x -> yield (a',b',c',d',e',x,g',h',i'));
+    g g' (fun x -> yield (a',b',c',d',e',f',x,h',i'));
+    h h' (fun x -> yield (a',b',c',d',e',f',g',x,i'));
+    i i' (fun x -> yield (a',b',c',d',e',f',g',h',x))
+
+  let tup9_opt a b c d e f g h i =
+    let a = Option.value ~default a in
+    let b = Option.value ~default b in
+    let c = Option.value ~default c in
+    let d = Option.value ~default d in
+    let e = Option.value ~default e in
+    let f = Option.value ~default f in
+    let g = Option.value ~default g in
+    let h = Option.value ~default h in
+    let i = Option.value ~default i in
+    tup9 a b c d e f g h i
 end
 
 (** {2 Observe Values} *)
@@ -638,6 +955,7 @@ module Observable = struct
     let int i = i land max_int
     let bool b = if b then 1 else 2
     let char x = Char.code x
+    let bytes (x:bytes) = Hashtbl.hash x
     let string (x:string) = Hashtbl.hash x
     let opt f = function
       | None -> 42
@@ -651,9 +969,10 @@ module Observable = struct
     type 'a t = 'a -> 'a -> bool
 
     let int : int t = (=)
+    let bytes : bytes t = (=)
     let string : string t = (=)
     let bool : bool t = (=)
-    let float : float t = (=)
+    let float = Float.equal
     let unit () () = true
     let char : char t = (=)
 
@@ -684,6 +1003,7 @@ module Observable = struct
   let bool : bool t = make ~hash:H.bool ~eq:Eq.bool Print.bool
   let int : int t = make ~hash:H.int ~eq:Eq.int Print.int
   let float : float t = make ~eq:Eq.float Print.float
+  let bytes = make ~hash:H.bytes ~eq:Eq.bytes Print.bytes
   let string = make ~hash:H.string ~eq:Eq.string Print.string
   let char = make ~hash:H.char ~eq:Eq.char Print.char
 
@@ -798,27 +1118,69 @@ let int64 =
   make ~print:(fun i -> Int64.to_string i ^ "L") ~small:small1
     ~shrink:Shrink.int64 Gen.ui64
 
-let char = make_scalar ~print:(sprintf "%C") Gen.char
-let printable_char = make_scalar ~print:(sprintf "%C") Gen.printable
-let numeral_char = make_scalar ~print:(sprintf "%C") Gen.numeral
+let small_char target c = abs ((Char.code c) - (Char.code target))
+
+let char =
+  make ~print:(sprintf "%C") ~small:(small_char 'a') ~shrink:Shrink.char Gen.char
+let printable_char =
+  make ~print:(sprintf "%C") ~small:(small_char 'a') ~shrink:Shrink.char_printable Gen.printable
+let numeral_char =
+  make ~print:(sprintf "%C") ~small:(small_char '0') ~shrink:Shrink.char_numeral Gen.numeral
+
+let bytes_gen_of_size size gen =
+  make ~shrink:Shrink.bytes ~small:Bytes.length
+    ~print:(Print.bytes) (Gen.bytes_size ~gen size)
+let bytes_of gen =
+  make ~shrink:Shrink.bytes ~small:Bytes.length
+    ~print:(Print.bytes) (Gen.bytes ~gen)
+
+let bytes = bytes_of Gen.char
+let bytes_of_size size = bytes_gen_of_size size Gen.char
+let bytes_small = bytes_gen_of_size Gen.small_nat Gen.char
+let bytes_small_of gen = bytes_gen_of_size Gen.small_nat gen
+let bytes_printable =
+  make ~shrink:(Shrink.bytes ~shrink:Shrink.char_printable) ~small:Bytes.length
+    ~print:(Print.bytes) (Gen.bytes ~gen:Gen.printable)
 
 let string_gen_of_size size gen =
   make ~shrink:Shrink.string ~small:String.length
     ~print:(sprintf "%S") (Gen.string_size ~gen size)
-let string_gen gen =
+let string_of gen =
   make ~shrink:Shrink.string ~small:String.length
     ~print:(sprintf "%S") (Gen.string ~gen)
 
-let string = string_gen Gen.char
+let string = string_of Gen.char
 let string_of_size size = string_gen_of_size size Gen.char
-let small_string = string_gen_of_size Gen.small_nat Gen.char
+let string_small = string_gen_of_size Gen.small_nat Gen.char
+let string_small_of gen = string_gen_of_size Gen.small_nat gen
+let small_string = string_small
+let string_gen = string_of
 
-let printable_string = string_gen Gen.printable
-let printable_string_of_size size = string_gen_of_size size Gen.printable
-let small_printable_string = string_gen_of_size Gen.small_nat Gen.printable
+let printable_string =
+  make ~shrink:(Shrink.string ~shrink:Shrink.char_printable) ~small:String.length
+    ~print:(sprintf "%S") (Gen.string ~gen:Gen.printable)
 
-let numeral_string = string_gen Gen.numeral
-let numeral_string_of_size size = string_gen_of_size size Gen.numeral
+let printable_string_of_size size =
+  make ~shrink:(Shrink.string ~shrink:Shrink.char_printable) ~small:String.length
+    ~print:(sprintf "%S") (Gen.string_size ~gen:Gen.printable size)
+
+let small_printable_string =
+  make ~shrink:(Shrink.string ~shrink:Shrink.char_printable) ~small:String.length
+    ~print:(sprintf "%S") (Gen.string_size ~gen:Gen.printable Gen.small_nat)
+
+let numeral_string =
+  make ~shrink:(Shrink.string ~shrink:Shrink.char_numeral) ~small:String.length
+    ~print:(sprintf "%S") (Gen.string ~gen:Gen.numeral)
+
+let numeral_string_of_size size =
+  make ~shrink:(Shrink.string ~shrink:Shrink.char_numeral) ~small:String.length
+    ~print:(sprintf "%S") (Gen.string_size ~gen:Gen.numeral size)
+
+let string_printable = printable_string
+let string_printable_of_size = printable_string_of_size
+let string_small_printable = small_printable_string
+let string_numeral = numeral_string
+let string_numeral_of_size = numeral_string_of_size
 
 let list_sum_ f l = List.fold_left (fun acc x-> f x+acc) 0 l
 
@@ -876,6 +1238,78 @@ let quad a b c d =
        (_opt_or d.shrink Shrink.nil))
     (Gen.quad a.gen b.gen c.gen d.gen)
 
+let tup2 a b=
+  make
+    ?small:(_opt_map_2 ~f:(fun a b (a', b') -> a a'+b b') a.small b.small)
+    ~print:(Print.tup2_opt a.print b.print)
+    ~shrink:(Shrink.pair (_opt_or a.shrink Shrink.nil) (_opt_or b.shrink Shrink.nil))
+    (Gen.tup2 a.gen b.gen)
+
+let tup3 a b c =
+  make
+    ?small:(_opt_map_3 ~f:(fun a b c (a', b', c') ->
+        a a'+b b'+c c') a.small b.small c.small)
+    ~print:(Print.tup3_opt a.print b.print c.print)
+    ~shrink:(Shrink.tup3_opt a.shrink b.shrink c.shrink)
+    (Gen.tup3 a.gen b.gen c.gen)
+
+let tup4 a b c d =
+  make
+    ?small:(_opt_map_4 ~f:(fun a b c d (a', b', c', d') ->
+        a a'+b b'+c c'+d d') a.small b.small c.small d.small)
+    ~print:(Print.tup4_opt a.print b.print c.print d.print)
+    ~shrink:(Shrink.tup4_opt a.shrink b.shrink c.shrink d.shrink)
+    (Gen.tup4 a.gen b.gen c.gen d.gen)
+
+let tup5 a b c d e =
+  make
+    ?small:(_opt_map_5 ~f:(fun a b c d e (a', b', c', d', e') ->
+        a a'+b b'+c c'+d d'+e e') a.small b.small c.small d.small e.small)
+    ~print:(Print.tup5_opt a.print b.print c.print d.print e.print)
+    ~shrink:(Shrink.tup5_opt a.shrink b.shrink c.shrink d.shrink e.shrink)
+    (Gen.tup5 a.gen b.gen c.gen d.gen e.gen)
+
+let tup6 a b c d e f =
+  make
+    ?small:(_opt_map_6 ~f:(fun a b c d e f (a', b', c', d', e', f') ->
+        a a'+b b'+c c'+d d'+e e'+f f') a.small b.small c.small d.small e.small f.small)
+    ~print:(Print.tup6_opt a.print b.print c.print d.print e.print f.print)
+    ~shrink:(Shrink.tup6_opt a.shrink b.shrink c.shrink d.shrink e.shrink f.shrink)
+    (Gen.tup6 a.gen b.gen c.gen d.gen e.gen f.gen)
+
+let tup7 a b c d e f g =
+  make
+    ?small:(_opt_map_7 ~f:(fun a b c d e f g (a', b', c', d', e', f', g') ->
+        a a'+b b'+c c'+d d'+e e'+f f'+g g')
+        a.small b.small c.small d.small e.small f.small g.small)
+    ~print:(Print.tup7_opt
+              a.print b.print c.print d.print e.print f.print g.print)
+    ~shrink:(Shrink.tup7_opt
+               a.shrink b.shrink c.shrink d.shrink e.shrink f.shrink g.shrink)
+    (Gen.tup7 a.gen b.gen c.gen d.gen e.gen f.gen g.gen)
+
+let tup8 a b c d e f g h =
+  make
+    ?small:(_opt_map_8 ~f:(fun a b c d e f g h (a', b', c', d', e', f', g', h') ->
+        a a'+b b'+c c'+d d'+e e'+f f'+g g'+h h')
+        a.small b.small c.small d.small e.small f.small g.small h.small)
+    ~print:(Print.tup8_opt
+              a.print b.print c.print d.print e.print f.print g.print h.print)
+    ~shrink:(Shrink.tup8_opt
+               a.shrink b.shrink c.shrink d.shrink e.shrink f.shrink g.shrink h.shrink)
+    (Gen.tup8 a.gen b.gen c.gen d.gen e.gen f.gen g.gen h.gen)
+
+let tup9 a b c d e f g h i =
+  make
+    ?small:(_opt_map_9 ~f:(fun a b c d e f g h i (a', b', c', d', e', f', g', h', i') ->
+        a a'+b b'+c c'+d d'+e e'+f f'+g g'+h h'+i i')
+        a.small b.small c.small d.small e.small f.small g.small h.small i.small)
+    ~print:(Print.tup9_opt
+              a.print b.print c.print d.print e.print f.print g.print h.print i.print)
+    ~shrink:(Shrink.tup9_opt
+               a.shrink b.shrink c.shrink d.shrink e.shrink f.shrink g.shrink h.shrink i.shrink)
+    (Gen.tup9 a.gen b.gen c.gen d.gen e.gen f.gen g.gen h.gen i.gen)
+
 let option ?ratio a =
   let g = Gen.opt ?ratio a.gen
   and shrink = _opt_map a.shrink ~f:Shrink.option
@@ -955,6 +1389,8 @@ end = struct
       List.iter (fun (k,v) -> T.add tbl k v) l;
       tbl
     in
+    (* split random state to avoid later failed [get]s to side-effect the current [st] *)
+    let st' = RS.split st in
     (* make a table
        @param extend if true, extend table on the fly *)
     let rec make ~extend tbl = {
@@ -962,7 +1398,7 @@ end = struct
         try Some (T.find tbl x)
         with Not_found ->
           if extend then (
-            let v = v.gen st in
+            let v = v.gen st' in
             T.add tbl x v;
             Some v
           ) else None);
@@ -977,7 +1413,7 @@ end = struct
             tbl;
         Buffer.contents b);
       p_shrink1=(fun yield ->
-        Shrink.list (tbl_to_list tbl)
+        Shrink.list_spine (tbl_to_list tbl)
           (fun l ->
              yield (make ~extend:false (tbl_of_list l)))
       );
@@ -1048,9 +1484,9 @@ module Fn = struct
       = function
         | Fun_tbl {fun_arb=a; fun_tbl=tbl; fun_default=def} ->
           let sh_v = match a.shrink with None -> Shrink.nil | Some s->s in
-          (Poly_tbl.shrink1 tbl >|= fun tbl' -> mk_repr tbl' a def)
-          <+>
             (sh_v def >|= fun def' -> mk_repr tbl a def')
+          <+>
+            (Poly_tbl.shrink1 tbl >|= fun tbl' -> mk_repr tbl' a def)
           <+>
             (Poly_tbl.shrink2 sh_v tbl >|= fun tbl' -> mk_repr tbl' a def)
         | Fun_map (g, r') ->
@@ -1328,14 +1764,17 @@ module Test = struct
   let get_long_factor = QCheck2.Test.get_long_factor
 
   let make_cell ?if_assumptions_fail
-      ?count ?long_factor ?max_gen
-  ?max_fail ?small:_removed_in_qcheck_2 ?name arb law
+      ?count ?long_factor ?negative ?max_gen
+  ?max_fail ?small:_removed_in_qcheck_2 ?retries ?name arb law
   =
   let {gen; shrink; print; collect; stats; _} = arb in
-  QCheck2.Test.make_cell_from_QCheck1 ?if_assumptions_fail ?count ?long_factor ?max_gen ?max_fail ?name ~gen ?shrink ?print ?collect ~stats law
+  QCheck2.Test.make_cell_from_QCheck1 ?if_assumptions_fail ?count ?long_factor ?negative ?max_gen ?max_fail ?retries ?name ~gen ?shrink ?print ?collect ~stats law
 
-  let make ?if_assumptions_fail ?count ?long_factor ?max_gen ?max_fail ?small ?name arb law =
-    QCheck2.Test.Test (make_cell ?if_assumptions_fail ?count ?long_factor ?max_gen ?max_fail ?small ?name arb law)
+  let make' ?if_assumptions_fail ?count ?long_factor ?max_gen ?max_fail ?small ?retries ?name ~negative arb law =
+    QCheck2.Test.Test (make_cell ?if_assumptions_fail ?count ?long_factor ?max_gen ?max_fail ?small ?retries ?name ~negative arb law)
+
+  let make = make' ~negative:false
+  let make_neg = make' ~negative:true
 
   let fail_report = QCheck2.Test.fail_report
 
