@@ -293,12 +293,13 @@ module Gen : sig
       @since 0.5.2 *)
 
   val int_bound : int -> int t
-  (** Uniform integer generator producing integers within [0... bound].
+  (** Uniform integer generator producing integers between [0] and [bound]
+      (inclusive).
       For [bound < 2^{30} - 1] uses [Random.State.int] for integer generation.
       @raise Invalid_argument if the argument is negative. *)
 
   val int_range : int -> int -> int t
-  (** Uniform integer generator producing integers within [low,high].
+  (** Uniform integer generator producing integers within [low,high] (inclusive).
       @raise Invalid_argument if [low > high]. *)
 
   val graft_corners : 'a t -> 'a list -> unit -> 'a t
@@ -314,7 +315,7 @@ module Gen : sig
   (** All corner cases for int.
       @since 0.6 *)
 
-  val (--) : int -> int -> int t (** Synonym to {!int_range}. *)
+  val (--) : int -> int -> int t (** Synonym for {!int_range}. *)
 
   val ui32 : int32 t (** Generates (unsigned) [int32] values. *)
 
@@ -338,10 +339,17 @@ module Gen : sig
   val array_repeat : int -> 'a t -> 'a array t
   (** [array_repeat i g] builds an array generator from exactly [i] elements generated by [g]. *)
 
-  val opt : ?ratio:float -> 'a t -> 'a option t
-  (**  An option generator, with optional ratio.
+  val option : ?ratio:float -> 'a t -> 'a option t
+  (** An option generator, with optional ratio.
       @param ratio a float between [0.] and [1.] indicating the probability of a sample to be [Some _]
       rather than [None].
+
+      @since 0.19 (renamed from [opt])
+  *)
+
+  val opt : ?ratio:float -> 'a t -> 'a option t
+  (** [opt] is an alias of {!val:option} for backward compatibility.
+
       @since 0.18 ([?ratio] parameter)
   *)
 
@@ -353,10 +361,34 @@ module Gen : sig
   (** Generates quadruples.
       @since 0.5.1 *)
 
+   (** {3 Tuple of generators} *)
+
+  (** {4 Shrinks on [gen1], then [gen2], then ... } *)
+
+  val tup2 : 'a t -> 'b t -> ('a * 'b) t
+
+  val tup3 : 'a t -> 'b t -> 'c t -> ('a * 'b * 'c) t
+
+  val tup4 : 'a t -> 'b t -> 'c t -> 'd t -> ('a * 'b * 'c * 'd) t
+
+  val tup5 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> ('a * 'b * 'c * 'd * 'e) t
+
+  val tup6 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t ->
+    ('a * 'b * 'c * 'd * 'e * 'f) t
+
+  val tup7 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t -> 'g t ->
+    ('a * 'b * 'c * 'd * 'e * 'f * 'g) t
+
+  val tup8 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t -> 'g t -> 'h t ->
+    ('a * 'b * 'c * 'd * 'e * 'f * 'g * 'h) t
+
+  val tup9 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t -> 'g t -> 'h t -> 'i t ->
+    ('a * 'b * 'c * 'd * 'e * 'f * 'g * 'h * 'i) t
+
   val char : char t
   (** Generates characters upto character code 255. *)
 
-  val printable : char t (** Generates printable characters. *)
+  val printable : char t (** Generates printable ascii characters - either '\n' or in the range 32 to 126, inclusive *)
 
   val numeral : char t (** Generates numeral characters. *)
 
@@ -365,6 +397,34 @@ module Gen : sig
       Example: [char_range 'a' 'z'] for all lower case ascii letters.
       @since 0.13 *)
 
+  val bytes_size : ?gen:char t -> int t -> bytes t
+  (** Builds a bytes generator from a (non-negative) size generator.
+      Accepts an optional character generator (the default is {!char}).
+      @since 0.20 *)
+
+  val bytes : ?gen:char t -> bytes t
+  (** Builds a bytes generator. Bytes size is generated by {!nat}.
+      Accepts an optional character generator (the default is {!char}).
+      See also {!bytes_of} and {!bytes_readable} for versions without
+      optional parameters.
+      @since 0.20 *)
+
+  val bytes_of : char t -> bytes t
+  (** Builds a bytes generator using the given character generator.
+      @since 0.20 *)
+
+  val bytes_printable : bytes t
+  (** Generator using the {!printable} character generator.
+      @since 0.20 *)
+
+  val bytes_small : bytes t
+  (** Builds a bytes generator using the {!char} character generator, length is {!small_nat}
+      @since 0.20 *)
+
+  val bytes_small_of : char t -> bytes t
+  (** Builds a bytes generator using the given character generator, length is {!small_nat}.
+      @since 0.20 *)
+
   val string_size : ?gen:char t -> int t -> string t
   (** Builds a string generator from a (non-negative) size generator.
       Accepts an optional character generator (the default is {!char}). *)
@@ -393,6 +453,14 @@ module Gen : sig
   (** Builds a string generator, length is {!small_nat}
       Accepts an optional character generator (the default is {!char}). *)
 
+  val string_small : string t
+  (** Builds a string generator using the {!char} character generator, length is {!small_nat}
+      @since 0.20 *)
+
+  val string_small_of : char t -> string t
+  (** Builds a string generator using the given character generator, length is {!small_nat}.
+      @since 0.20 *)
+
   val small_list : 'a t -> 'a list t
   (** Generates lists of small size (see {!small_nat}).
       @since 0.5.3 *)
@@ -545,6 +613,8 @@ module Print : sig
 
   val char : char t (** Character printer. *)
 
+  val bytes : bytes t (** Bytes printer. @since 0.20 *)
+
   val string : string t (** String printer. *)
 
   val option : 'a t -> 'a option t (** Option printer. *)
@@ -567,6 +637,34 @@ module Print : sig
   val comap : ('a -> 'b) -> 'b t -> 'a t
   (** [comap f p] maps [p], a printer of type ['b], to a printer of type ['a] by
       first converting a printed value using [f : 'a -> 'b]. *)
+
+  val tup2 : 'a t -> 'b t -> ('a * 'b) t
+  (** 2-tuple printer. Expects printers for each component. *)
+
+  val tup3 : 'a t -> 'b t -> 'c t -> ('a * 'b * 'c) t
+  (** 3-tuple printer. Expects printers for each component. *)
+
+  val tup4 : 'a t -> 'b t -> 'c t -> 'd t -> ('a * 'b * 'c * 'd) t
+  (** 4-tuple printer. Expects printers for each component. *)
+
+  val tup5 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> ('a * 'b * 'c * 'd * 'e) t
+  (** 5-tuple printer. Expects printers for each component. *)
+
+  val tup6 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t ->
+    ('a * 'b * 'c * 'd * 'e * 'f) t
+  (** 6-tuple printer. Expects printers for each component. *)
+
+  val tup7 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t -> 'g t ->
+    ('a * 'b * 'c * 'd * 'e * 'f * 'g) t
+  (** 7-tuple printer. Expects printers for each component. *)
+
+  val tup8 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t -> 'g t -> 'h t ->
+    ('a * 'b * 'c * 'd * 'e * 'f * 'g * 'h) t
+  (** 8-tuple printer. Expects printers for each component. *)
+
+  val tup9 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t -> 'g t -> 'h t -> 'i t ->
+    ('a * 'b * 'c * 'd * 'e * 'f * 'g * 'h * 'i) t
+  (** 9-tuple printer. Expects printers for each component. *)
 end
 
 (** {2 Iterators}
@@ -585,7 +683,7 @@ module Iter : sig
   val map2 : ('a -> 'b -> 'c) -> 'a t -> 'b t -> 'c t
   val (>|=) : 'a t -> ('a -> 'b) -> 'b t
   val append : 'a t -> 'a t -> 'a t
-  val (<+>) : 'a t -> 'a t -> 'a t (** Synonym to {!append}. *)
+  val (<+>) : 'a t -> 'a t -> 'a t (** Synonym for {!append}. *)
 
   val of_list : 'a list -> 'a t
   val of_array : 'a array -> 'a t
@@ -627,7 +725,17 @@ module Shrink : sig
 
   val unit : unit t (** @since 0.6 *)
 
-  val char : char t (** @since 0.6 *)
+  val char : char t
+  (** Shrinks towards ['a'].
+      @since 0.6 *)
+
+  val char_numeral : char t
+  (** Shrinks towards ['0'].
+      @since 0.19 *)
+
+  val char_printable : char t
+  (** Shrinks towards ['a'] like [!char]. The output is also a printable character.
+      @since 0.19 *)
 
   val int : int t
 
@@ -639,7 +747,10 @@ module Shrink : sig
 
   val option : 'a t -> 'a option t
 
-  val string : string t
+  val bytes : ?shrink:(char t) -> bytes t
+  (** @since 0.20 *)
+
+  val string : ?shrink:(char t) -> string t
 
   val filter : ('a -> bool) -> 'a t -> 'a t
   (** [filter f shrink] shrinks values the same as [shrink], but
@@ -682,6 +793,38 @@ module Shrink : sig
 
   val quad : 'a t -> 'b t -> 'c t -> 'd t -> ('a * 'b * 'c * 'd) t
   (** Similar to {!pair} *)
+
+  val tup2 : 'a t -> 'b t -> ('a * 'b) t
+  (** [tup2 a b] uses [a] to shrink the first element of tuples,
+      then tries to shrink the second element using [b].
+      It is often better, when generating tuples, to put the "simplest"
+      element first (atomic type rather than list, etc.) because it will be
+      shrunk earlier. In particular, putting functions last might help. *)
+
+  val tup3 : 'a t -> 'b t -> 'c t -> ('a * 'b * 'c) t
+  (** Similar to {!tup2} *)
+
+  val tup4 : 'a t -> 'b t -> 'c t -> 'd t -> ('a * 'b * 'c * 'd) t
+  (** Similar to {!tup2} *)
+
+  val tup5 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> ('a * 'b * 'c * 'd * 'e) t
+  (** Similar to {!tup2} *)
+
+  val tup6 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t ->
+    ('a * 'b * 'c * 'd * 'e * 'f) t
+  (** Similar to {!tup2} *)
+
+  val tup7 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t -> 'g t ->
+    ('a * 'b * 'c * 'd * 'e * 'f * 'g) t
+  (** Similar to {!tup2} *)
+
+  val tup8 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t -> 'g t -> 'h t ->
+    ('a * 'b * 'c * 'd * 'e * 'f * 'g * 'h) t
+  (** Similar to {!tup2} *)
+
+  val tup9 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t -> 'g t -> 'h t -> 'i t ->
+    ('a * 'b * 'c * 'd * 'e * 'f * 'g * 'h * 'i) t
+  (** Similar to {!tup2} *)
 end
 
 (** {2 Observe Values} *)
@@ -709,6 +852,7 @@ module Observable : sig
   val int : int t
   val float : float t
   val string : string t
+  val bytes : bytes t (** @since 0.20 *)
   val char : char t
 
   val make :
@@ -806,10 +950,17 @@ val get_print : 'a arbitrary -> 'a Print.t option
     with an object of type [foo arbitrary] used to generate, print, etc. values
     of type [foo].
 
-    See {!Test.make} to build a test, and {!Test.check_exn} to
-    run one test simply.
-    For more serious testing, it is better to create a testsuite
-    and use {!QCheck_runner}.
+    The main features of this module are:
+    - {!Test.make} to build a test,
+    - {!Test.make_neg} to build a negative test that is expected not to satisfy the tested property,
+    - {!Test.check_exn} to run a single test with a simple runner.
+
+    A test fails if the property does not hold for a given input. The {{!Test.fail_report} simple} form or the {{!Test.fail_reportf} rich} form) offer more elaborate forms to fail a test.
+
+    For more serious testing, it is recommended to create a testsuite and use a full-fledged runner:
+    - {!QCheck_base_runner} is a QCheck-only runner (useful if you don't have or don't need another test framework)
+    - {!QCheck_alcotest} interfaces to the Alcotest framework
+    - {!QCheck_ounit} interfaces to the to OUnit framework
 *)
 
 (** Result of running a test *)
@@ -903,16 +1054,18 @@ module Test : sig
 
   val make_cell :
     ?if_assumptions_fail:([`Fatal | `Warning] * float) ->
-    ?count:int -> ?long_factor:int -> ?max_gen:int -> ?max_fail:int ->
-    ?small:('a -> int) -> ?name:string -> 'a arbitrary -> ('a -> bool) ->
-    'a cell
+    ?count:int -> ?long_factor:int -> ?negative:bool -> ?max_gen:int -> ?max_fail:int ->
+    ?small:('a -> int) -> ?retries:int -> ?name:string ->
+    'a arbitrary -> ('a -> bool) -> 'a cell
   (** [make_cell arb prop] builds a test that checks property [prop] on instances
       of the generator [arb].
       @param name the name of the test.
       @param count number of test cases to run, counting only
         the test cases which satisfy preconditions.
+      @param retries number of times to retry the tested property while shrinking.
       @param long_factor the factor by which to multiply count, max_gen and
         max_fail when running a long test (default: 1).
+      @param negative whether the test is expected not to satisfy the tested property.
       @param max_gen maximum number of times the generation function
         is called in total to replace inputs that do not satisfy
         preconditions (should be >= count).
@@ -951,12 +1104,25 @@ module Test : sig
   val make :
     ?if_assumptions_fail:([`Fatal | `Warning] * float) ->
     ?count:int -> ?long_factor:int -> ?max_gen:int -> ?max_fail:int ->
-    ?small:('a -> int) -> ?name:string -> 'a arbitrary -> ('a -> bool) -> t
+    ?small:('a -> int) -> ?retries:int -> ?name:string -> 'a arbitrary ->
+    ('a -> bool) -> t
   (** [make arb prop] builds a test that checks property [prop] on instances
       of the generator [arb].
       See {!make_cell} for a description of the parameters.
   *)
 
+  val make_neg :
+    ?if_assumptions_fail:([`Fatal | `Warning] * float) ->
+    ?count:int -> ?long_factor:int -> ?max_gen:int -> ?max_fail:int ->
+    ?small:('a -> int) -> ?retries:int -> ?name:string -> 'a arbitrary ->
+    ('a -> bool) -> t
+  (** [make_neg arb prop] builds a test that checks property [prop] on instances
+      of the generator [arb].
+      The test is considered negative, meaning that it is expected not to satisfy the tested property.
+      This information is recorded in an underlying test [cell] entry and interpreted suitably by test runners.
+      See {!make_cell} for a description of the parameters.
+  *)
+
   include module type of QCheck2.Test_exceptions
 
   val print_instance : 'a cell -> 'a -> string
@@ -973,7 +1139,8 @@ module Test : sig
     ?rand:Random.State.t -> 'a cell -> 'a TestResult.t
 
   val check_cell_exn :
-    ?long:bool -> ?call:'a callback -> ?step:'a step ->
+    ?long:bool -> ?call:'a callback ->
+    ?step:'a step -> ?handler:'a handler ->
     ?rand:Random.State.t -> 'a cell -> unit
 
   val check_exn : ?long:bool -> ?rand:Random.State.t -> t -> unit
@@ -1082,7 +1249,7 @@ val small_signed_int : int arbitrary
     @since 0.5.2 *)
 
 val (--) : int -> int -> int arbitrary
-(** Synonym to {!int_range}. *)
+(** Synonym for {!int_range}. *)
 
 val int32 : int32 arbitrary
 (** Int32 generator. Uniformly distributed. *)
@@ -1109,55 +1276,121 @@ val char : char arbitrary
     valid latin-1). *)
 
 val printable_char : char arbitrary
-(** Uniformly distributed over a subset of chars. *)
-(* FIXME: describe which subset. *)
+(** Uniformly distributed over a subset of printable ascii chars.
+    Ascii character codes 32 to 126, inclusive - or ['\n'] with code 10.
+  *)
 
 val numeral_char : char arbitrary
 (** Uniformly distributed over ['0'..'9']. *)
 
+val bytes_gen_of_size : int Gen.t -> char Gen.t -> bytes arbitrary
+(** Builds a bytes generator from a (non-negative) size generator and a character generator.
+    @since 0.20 *)
+
+val bytes_of : char Gen.t -> bytes arbitrary
+(** Generates bytes with a distribution of length of {!Gen.nat}.
+    @since 0.20 *)
+
+val bytes : bytes arbitrary
+(** Generates bytes with a distribution of length of {!Gen.nat}
+    and distribution of characters of [char].
+    @since 0.20 *)
+
+val bytes_small : bytes arbitrary
+(** Same as {!bytes} but with a small length (ie {!Gen.small_nat} ).
+    @since 0.20 *)
+
+val bytes_small_of : char Gen.t -> bytes arbitrary
+(** Same as {!bytes_of} but with a small length (ie {!Gen.small_nat} ).
+    @since 0.20 *)
+
+val bytes_of_size : int Gen.t -> bytes arbitrary
+(** Generates bytes with distribution of characters of [char].
+    @since 0.20 *)
+
+val bytes_printable : bytes arbitrary
+(** Generates bytes with a distribution of length of {!Gen.nat}
+    and distribution of characters of [printable_char].
+    @since 0.20 *)
+
 val string_gen_of_size : int Gen.t -> char Gen.t -> string arbitrary
+(** Builds a string generator from a (non-negative) size generator and a character generator. *)
 
 val string_gen : char Gen.t -> string arbitrary
-(** Generates strings with a distribution of length of [small_nat]. *)
+(** Generates strings with a distribution of length of {!Gen.nat}. *)
+
+val string_of : char Gen.t -> string arbitrary
+(** Synonym for {!string_gen} added for convenience.
+    @since 0.20 *)
 
 val string : string arbitrary
-(** Generates strings with a distribution of length of [small_nat]
+(** Generates strings with a distribution of length of {!Gen.nat}
     and distribution of characters of [char]. *)
 
 val small_string : string arbitrary
 (** Same as {!string} but with a small length (ie {!Gen.small_nat} ). *)
 
+val string_small : string arbitrary
+(** Synonym for [small_string] added for convenience.
+    @since 0.20 *)
+
+val string_small_of : char Gen.t -> string arbitrary
+(** Same as {!string_of} but with a small length (ie {!Gen.small_nat} ).
+    @since 0.20 *)
+
 val small_list : 'a arbitrary -> 'a list arbitrary
 (** Generates lists of small size (see {!Gen.small_nat}).
     @since 0.5.3 *)
 
 val string_of_size : int Gen.t -> string arbitrary
-(** Generates strings with distribution of characters if [char]. *)
+(** Generates strings with distribution of characters of [char]. *)
 
 val printable_string : string arbitrary
-(** Generates strings with a distribution of length of [small_nat]
+(** Generates strings with a distribution of length of {!Gen.nat}
     and distribution of characters of [printable_char]. *)
 
+val string_printable : string arbitrary
+(** Synonym for [printable_string] added for convenience.
+    @since 0.20 *)
+
 val printable_string_of_size : int Gen.t -> string arbitrary
 (** Generates strings with distribution of characters of [printable_char]. *)
 
+val string_printable_of_size : int Gen.t -> string arbitrary
+(** Synonym for [printable_string_of_size] added for convenience.
+    @since 0.20 *)
+
 val small_printable_string : string arbitrary
+(** Generates strings with a length of [small_nat]
+    and distribution of characters of [printable_char]. *)
+
+val string_small_printable : string arbitrary
+(** Synonym for [small_printable_string] added for convenience.
+    @since 0.20 *)
 
 val numeral_string : string arbitrary
-(** Generates strings with a distribution of length of [small_nat]
+(** Generates strings with a distribution of length of {!Gen.nat}
     and distribution of characters of [numeral_char]. *)
 
+val string_numeral : string arbitrary
+(** Synonym for [numeral_string] added for convenience.
+    @since 0.20 *)
+
 val numeral_string_of_size : int Gen.t -> string arbitrary
 (** Generates strings with a distribution of characters of [numeral_char]. *)
 
+val string_numeral_of_size : int Gen.t -> string arbitrary
+(** Synonym for [numeral_string_of_size] added for convenience.
+    @since 0.20 *)
+
 val list : 'a arbitrary -> 'a list arbitrary
-(** Generates lists with length generated by [small_nat]. *)
+(** Generates lists with length generated by {!Gen.nat}. *)
 
 val list_of_size : int Gen.t -> 'a arbitrary -> 'a list arbitrary
 (** Generates lists with length from the given distribution. *)
 
 val array : 'a arbitrary -> 'a array arbitrary
-(** Generates arrays with length generated by [small_nat]. *)
+(** Generates arrays with length generated by {!Gen.nat}. *)
 
 val array_of_size : int Gen.t -> 'a arbitrary -> 'a array arbitrary
 (** Generates arrays with length from the given distribution. *)
@@ -1174,13 +1407,108 @@ val quad : 'a arbitrary -> 'b arbitrary -> 'c arbitrary -> 'd arbitrary -> ('a *
 (** Combines four generators into a generator of 4-tuples.
     Order matters for shrinking, see {!Shrink.pair} and the likes *)
 
+(** {3 Tuple of generators} *)
+
+(** {4 Shrinks on [gen1], then [gen2], then ... } *)
+
+val tup2 :
+  'a arbitrary ->
+  'b arbitrary ->
+  ('a * 'b) arbitrary
+(** Combines two generators into a 2-tuple generator.
+    Order of elements can matter (w.r.t shrinking, see {!Shrink.tup2})
+    Prints as many elements as available printers *)
+
+val tup3 :
+  'a arbitrary ->
+  'b arbitrary ->
+  'c arbitrary ->
+  ('a * 'b * 'c) arbitrary
+(** Combines three generators into a 3-tuple generator.
+    Order of elements can matter (w.r.t shrinking, see {!Shrink.tup2})
+    Prints as many elements as available printers *)
+
+val tup4 :
+  'a arbitrary ->
+  'b arbitrary ->
+  'c arbitrary ->
+  'd arbitrary ->
+  ('a * 'b * 'c * 'd) arbitrary
+(** Combines four generators into a 4-tuple generator.
+    Order of elements can matter (w.r.t shrinking, see {!Shrink.tup2})
+    Prints as many elements as available printers *)
+
+val tup5 : 'a arbitrary ->
+  'b arbitrary ->
+  'c arbitrary ->
+  'd arbitrary ->
+  'e arbitrary ->
+  ('a * 'b * 'c * 'd * 'e) arbitrary
+(** Combines five generators into a 5-tuple generator.
+    Order of elements can matter (w.r.t shrinking, see {!Shrink.tup2})
+    Prints as many elements as available printers *)
+
+val tup6 :
+  'a arbitrary ->
+  'b arbitrary ->
+  'c arbitrary ->
+  'd arbitrary ->
+  'e arbitrary ->
+  'f arbitrary ->
+  ('a * 'b * 'c * 'd * 'e * 'f) arbitrary
+(** Combines six generators into a 6-tuple generator.
+    Order of elements can matter (w.r.t shrinking, see {!Shrink.tup2})
+    Prints as many elements as available printers *)
+
+val tup7 :
+  'a arbitrary ->
+  'b arbitrary ->
+  'c arbitrary ->
+  'd arbitrary ->
+  'e arbitrary ->
+  'f arbitrary ->
+  'g arbitrary ->
+  ('a * 'b * 'c * 'd * 'e * 'f * 'g) arbitrary
+(** Combines seven generators into a 7-tuple generator.
+    Order of elements can matter (w.r.t shrinking, see {!Shrink.tup2})
+    Prints as many elements as available printers *)
+
+val tup8 :
+  'a arbitrary ->
+  'b arbitrary ->
+  'c arbitrary ->
+  'd arbitrary ->
+  'e arbitrary ->
+  'f arbitrary ->
+  'g arbitrary ->
+  'h arbitrary ->
+  ('a * 'b * 'c * 'd * 'e * 'f * 'g * 'h) arbitrary
+(** Combines eight generators into a 8-tuple generator.
+    Order of elements can matter (w.r.t shrinking, see {!Shrink.tup2})
+    Prints as many elements as available printers *)
+
+val tup9 :
+  'a arbitrary ->
+  'b arbitrary ->
+  'c arbitrary ->
+  'd arbitrary ->
+  'e arbitrary ->
+  'f arbitrary ->
+  'g arbitrary ->
+  'h arbitrary ->
+  'i arbitrary ->
+  ('a * 'b * 'c * 'd * 'e * 'f * 'g * 'h * 'i) arbitrary
+(** Combines nine generators into a 9-tuple generator.
+    Order of elements can matter (w.r.t shrinking, see {!Shrink.tup2})
+    Prints as many elements as available printers *)
+
 val option : ?ratio:float -> 'a arbitrary -> 'a option arbitrary
 (** Choose between returning Some random value with optional ratio, or None. *)
 
 val fun1_unsafe : 'a arbitrary -> 'b arbitrary -> ('a -> 'b) arbitrary
 (** Generator of functions of arity 1.
     The functions are always pure and total functions:
-    - when given the same argument (as decided by Pervasives.(=)), it returns the same value
+    - when given the same argument (as decided by Stdlib.(=)), it returns the same value
     - it never does side effects, like printing or never raise exceptions etc.
       The functions generated are really printable.
 
