@@ -280,7 +280,7 @@ module Gen : sig
     - [32] to [126], inclusive
     - ['\n']
 
-    Shrinks towards ['a'].
+    Shrinks towards ['a'] or lower character codes.
   *)
 
   val numeral : char t
@@ -289,6 +289,51 @@ module Gen : sig
       Shrinks towards ['0'].
   *)
 
+  val bytes_size : ?gen:char t -> int t -> bytes t
+  (** Builds a bytes generator from a (non-negative) size generator.
+      Accepts an optional character generator (the default is {!char}).
+
+      Shrinks on the number of characters first, then on the characters.
+
+      @since 0.20 *)
+
+  val bytes : bytes t
+  (** Bytes generator using the {!char} character generator. Bytes size is generated by {!nat}.
+      See also {!bytes_of} and {!bytes_printable} for versions with
+      custom char generator.
+
+      Shrinks on the number of characters first, then on the characters.
+
+      @since 0.20 *)
+
+  val bytes_of : char t -> bytes t
+  (** Builds a bytes generator using the given character generator.
+
+      Shrinks on the number of characters first, then on the characters.
+
+      @since 0.20 *)
+
+  val bytes_printable : bytes t
+  (** Generator using the {!printable} character generator.
+
+      Shrinks on the number of characters first, then on the characters.
+
+      @since 0.20 *)
+
+  val bytes_small : bytes t
+  (** Builds a bytes generator using the {!char} character generator, length is {!small_nat}.
+
+      Shrinks on the number of characters first, then on the characters.
+
+      @since 0.20 *)
+
+  val bytes_small_of : char t -> bytes t
+  (** Builds a bytes generator using the given character generator, length is {!small_nat}.
+
+      Shrinks on the number of characters first, then on the characters.
+
+      @since 0.20 *)
+
   val string_size : ?gen:char t -> int t -> string t
   (** Builds a string generator from a (non-negative) size generator.
       Accepts an optional character generator (the default is {!char}).
@@ -319,11 +364,29 @@ module Gen : sig
 
       @since 0.11 *)
 
+  val string_small : string t
+  (** Builds a string generator using the {!char} characher generator, length is {!small_nat}.
+
+      Shrinks on the number of characters first, then on the characters.
+
+      @since 0.20
+  *)
+
+  val string_small_of : char t -> string t
+  (** Builds a string generator using the given characher generator, length is {!small_nat}.
+
+      Shrinks on the number of characters first, then on the characters.
+
+      @since 0.20
+  *)
+
   val small_string : ?gen:char t -> string t
   (** Builds a string generator, length is {!small_nat}.
       Accepts an optional character generator (the default is {!char}).
-
       Shrinks on the number of characters first, then on the characters.
+      This function is kept for backward compatibility:
+      The optional argument is in fact a mandatory [option], see c-cube/qcheck#162.
+      Use {!string_small} instead.
   *)
 
   val pure : 'a -> 'a t
@@ -373,6 +436,13 @@ module Gen : sig
       keep it and wait for users feedback (hence deprecation to raise attention).
   *)
 
+  val set_shrink : ('a -> 'a Seq.t) -> 'a t -> 'a t
+  (** [set_shrink shrink gen] sets the shrinker to [shrink] for [gen]. *)
+
+  val no_shrink : 'a t -> 'a t
+  (** [no_shrink gen] returns a generator using [gen] but with shrinking
+      disabled *)
+
   (** {3 Ranges} *)
 
   val int_bound : int -> int t
@@ -595,15 +665,20 @@ module Gen : sig
       Shrinks on elements only.
   *)
 
-  val opt : ?ratio:float -> 'a t -> 'a option t
-  (** [opt gen] is an [option] generator that uses [gen] when generating [Some] values.
+  val option : ?ratio:float -> 'a t -> 'a option t
+  (** [option gen] is an [option] generator that uses [gen] when generating [Some] values.
 
       Shrinks towards {!None} then towards shrinks of [gen].
 
       @param ratio a float between [0.] and [1.] indicating the probability of a sample to be [Some _]
       rather than [None] (value is [0.85]).
+
+      @since 0.19 (renamed from [opt])
   *)
 
+  val opt : ?ratio:float -> 'a t -> 'a option t
+  (** [opt] is an alias of {!val:option} for backward compatibility. *)
+
   (** {3 Combining generators} *)
 
   val pair : 'a t -> 'b t -> ('a * 'b) t
@@ -626,6 +701,26 @@ module Gen : sig
       @since 0.5.1
   *)
 
+  (** {3 Tuple of generators} *)
+
+  (** {4 Shrinks on [gen1], then [gen2], then ... } *)
+
+  val tup2 : 'a t -> 'b t -> ('a * 'b) t
+
+  val tup3 : 'a t -> 'b t -> 'c t -> ('a * 'b * 'c) t
+
+  val tup4 : 'a t -> 'b t -> 'c t -> 'd t -> ('a * 'b * 'c * 'd) t
+
+  val tup5 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> ('a * 'b * 'c * 'd * 'e) t
+
+  val tup6 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t -> ('a * 'b * 'c * 'd * 'e * 'f) t
+
+  val tup7 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t -> 'g t -> ('a * 'b * 'c * 'd * 'e * 'f * 'g) t
+
+  val tup8 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t -> 'g t -> 'h t -> ('a * 'b * 'c * 'd * 'e * 'f * 'g * 'h) t
+
+  val tup9 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t -> 'g t -> 'h t -> 'i t -> ('a * 'b * 'c * 'd * 'e * 'f * 'g * 'h * 'i) t
+
   (** {3 Convert a structure of generator to a generator of structure} *)
 
   val flatten_l : 'a t list -> 'a list t
@@ -1004,6 +1099,10 @@ module Print : sig
   val char : char t
   (** [char] is a printer of character. *)
 
+  val bytes : bytes t
+  (** [bytes] is a printer of bytes.
+      @since 0.20 *)
+
   val string : string t
   (** [string] is a printer of string. *)
 
@@ -1036,6 +1135,34 @@ module Print : sig
 
   val comap : ('b -> 'a) -> 'a t -> 'b t
   (** @deprecated use {!contramap} instead. *)
+
+  val tup2 : 'a t -> 'b t -> ('a * 'b) t
+  (** 2-tuple printer. Expects printers for each component. *)
+
+  val tup3 : 'a t -> 'b t -> 'c t -> ('a * 'b * 'c) t
+  (** 3-tuple printer. Expects printers for each component. *)
+
+  val tup4 : 'a t -> 'b t -> 'c t -> 'd t -> ('a * 'b * 'c * 'd) t
+  (** 4-tuple printer. Expects printers for each component. *)
+
+  val tup5 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> ('a * 'b * 'c * 'd * 'e) t
+  (** 5-tuple printer. Expects printers for each component. *)
+
+  val tup6 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t ->
+    ('a * 'b * 'c * 'd * 'e * 'f) t
+  (** 6-tuple printer. Expects printers for each component. *)
+
+  val tup7 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t -> 'g t ->
+    ('a * 'b * 'c * 'd * 'e * 'f * 'g) t
+  (** 7-tuple printer. Expects printers for each component. *)
+
+  val tup8 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t -> 'g t -> 'h t ->
+    ('a * 'b * 'c * 'd * 'e * 'f * 'g * 'h) t
+  (** 8-tuple printer. Expects printers for each component. *)
+
+  val tup9 : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t -> 'g t -> 'h t -> 'i t ->
+    ('a * 'b * 'c * 'd * 'e * 'f * 'g * 'h * 'i) t
+  (** 9-tuple printer. Expects printers for each component. *)
 end
 
 (** Shrinking helper functions. *)
@@ -1174,6 +1301,10 @@ module Observable : sig
   val float : float t
   (** [float] is an observable of [float]. *)
 
+  val bytes : bytes t
+  (** [bytes] is an observable of [bytes].
+      @since 0.20 *)
+
   val string : string t
   (** [string] is an observable of [string]. *)
 
@@ -1214,7 +1345,7 @@ module Observable : sig
   (** [quad o1 o2 o3 o4] is an observable of quadruples of [('a * 'b * 'c * 'd)]. *)
 end
 
-  
+
 (** Utils on combining function arguments. *)
 module Tuple : sig
   (** Heterogeneous tuple, used to pass any number of arguments to
@@ -1484,9 +1615,12 @@ module TestResult : sig
       @since 0.18 *)
 
   val is_success : _ t -> bool
-  (** Returns true iff the state if [Success]
+  (** Returns true iff the state is [Success]
       @since 0.9 *)
 
+  val is_failed : _ t -> bool
+  (** Returns true iff the state is [Failed _] *)
+
   val stats : 'a t -> ('a stat * (int,int) Hashtbl.t) list
   (** Obtain statistics
       @since 0.6
@@ -1517,12 +1651,14 @@ module Test_exceptions : sig
       stacktrace (if enabled) or an empty string. *)
 end
 
-(** A test is a pair of an generator and a property thar all generated values must satisfy. *)
+(** A test is a pair of a generator and a property that all generated values must satisfy. *)
 module Test : sig
   (** The main features of this module are:
-      - {!make} a test
-      - fail the test if a property does not hold (using either the {{!fail_report} simple} form or the {{!fail_reportf} rich} form)
-      - {!check_exn} a single test
+      - {!make} to create a test
+      - {!make_neg} to create a negative test that is expected not to satisfy the tested property
+      - {!check_exn} to run a single test with a simple runner.
+
+      A test fails if the property does not hold for a given input. The {{!fail_report} simple} form or the {{!fail_reportf} rich} form) offer more elaborate forms to fail a test.
 
       Note that while {!check_exn} is provided for convenience to discover QCheck or to run a single test in {{: https://opam.ocaml.org/blog/about-utop/} utop}, to run QCheck tests in your project you probably want to opt for a more advanced runner, or convert
       QCheck tests to your favorite test framework:
@@ -1534,11 +1670,11 @@ module Test : sig
   type 'a cell
   (** A single property test on a value of type ['a]. A {!Test.t} wraps a [cell]
       and hides its type parameter. *)
-     
+
   val make_cell :
     ?if_assumptions_fail:([`Fatal | `Warning] * float) ->
-    ?count:int -> ?long_factor:int -> ?max_gen:int -> ?max_fail:int -> ?name:string ->
-    ?print:'a Print.t -> ?collect:('a -> string) -> ?stats:('a stat list) ->
+    ?count:int -> ?long_factor:int ->  ?negative:bool -> ?max_gen:int -> ?max_fail:int -> ?retries:int ->
+    ?name:string -> ?print:'a Print.t -> ?collect:('a -> string) -> ?stats:('a stat list) ->
      'a Gen.t -> ('a -> bool) ->
     'a cell
   (** [make_cell gen prop] builds a test that checks property [prop] on instances
@@ -1548,11 +1684,13 @@ module Test : sig
         the test cases which satisfy preconditions.
       @param long_factor the factor by which to multiply count, max_gen and
         max_fail when running a long test (default: 1).
+      @param negative whether the test is expected not to satisfy the tested property.
       @param max_gen maximum number of times the generation function
         is called in total to replace inputs that do not satisfy
         preconditions (should be >= count).
       @param max_fail maximum number of failures before we stop generating
         inputs. This is useful if shrinking takes too much time.
+      @param retries number of times to retry the tested property while shrinking.
       @param if_assumptions_fail the minimum
         fraction of tests that must satisfy the precondition for a success
         to be considered valid.
@@ -1567,11 +1705,11 @@ module Test : sig
 
   val make_cell_from_QCheck1 :
     ?if_assumptions_fail:([`Fatal | `Warning] * float) ->
-    ?count:int -> ?long_factor:int -> ?max_gen:int -> ?max_fail:int ->
-    ?name:string -> gen:(Random.State.t -> 'a) -> ?shrink:('a -> ('a -> unit) -> unit) ->
+    ?count:int -> ?long_factor:int -> ?negative:bool -> ?max_gen:int -> ?max_fail:int ->
+    ?retries:int -> ?name:string -> gen:(Random.State.t -> 'a) -> ?shrink:('a -> ('a -> unit) -> unit) ->
     ?print:('a -> string) -> ?collect:('a -> string) -> stats:'a stat list -> ('a -> bool) ->
     'a cell
-  (** ⚠️ Do not use, this is exposed for internal reasons only. ⚠️ 
+  (** ⚠️ Do not use, this is exposed for internal reasons only. ⚠️
 
       @deprecated Migrate to QCheck2 and use {!make_cell} instead.
    *)
@@ -1592,22 +1730,39 @@ module Test : sig
   (** Get the long factor of a cell.
       @since 0.5.3 *)
 
+  val get_positive : _ cell -> bool
+  (** Get the expected mode of a cell: positive indicates expected to satisfy the tested property, negative indicates expected not to satisfy the tested property.  *)
+
   type t = Test : 'a cell -> t
   (** Same as ['a cell], but masking the type parameter. This allows to
       put tests on different types in the same list of tests. *)
 
   val make :
     ?if_assumptions_fail:([`Fatal | `Warning] * float) ->
-    ?count:int -> ?long_factor:int -> ?max_gen:int -> ?max_fail:int -> ?name:string ->
-    ?print:('a Print.t) -> ?collect:('a -> string) -> ?stats:('a stat list) ->
+    ?count:int -> ?long_factor:int -> ?max_gen:int -> ?max_fail:int -> ?retries:int ->
+    ?name:string -> ?print:('a Print.t) -> ?collect:('a -> string) -> ?stats:('a stat list) ->
     'a Gen.t -> ('a -> bool) -> t
   (** [make gen prop] builds a test that checks property [prop] on instances
       of the generator [gen].
       See {!make_cell} for a description of the parameters.
   *)
 
+  val make_neg :
+    ?if_assumptions_fail:([`Fatal | `Warning] * float) ->
+    ?count:int -> ?long_factor:int -> ?max_gen:int -> ?max_fail:int -> ?retries:int ->
+    ?name:string -> ?print:('a Print.t) -> ?collect:('a -> string) -> ?stats:('a stat list) ->
+    'a Gen.t -> ('a -> bool) -> t
+  (** [make_neg gen prop] builds a test that checks property [prop] on instances
+      of the generator [gen].
+      The test is considered negative, meaning that it is expected not to satisfy the tested property.
+      This information is recorded in an underlying test [cell] entry and interpreted suitably by test runners.
+      See {!make_cell} for a description of the parameters.
+  *)
+
   val test_get_count : t -> int
 
+  val test_get_long_factor : t -> int
+
   val fail_report : string -> 'a
   (** Fail the test with some additional message that will be reported.
 
@@ -1632,6 +1787,7 @@ module Test : sig
   val print_c_ex : 'a cell -> 'a TestResult.counter_ex -> string
   val print_fail : 'a cell -> string -> 'a TestResult.counter_ex list -> string
   val print_fail_other : string -> msg:string -> string
+  val print_expected_failure : 'a cell -> 'a TestResult.counter_ex list -> string
   val print_error : ?st:string -> 'a cell -> string -> 'a TestResult.counter_ex * exn -> string
   val print_test_fail : string -> string list -> string
   val print_test_error : string -> string -> exn -> string -> string
@@ -1647,8 +1803,8 @@ module Test : sig
   val check_result : 'a cell -> 'a TestResult.t -> unit
   (** [check_result cell res] checks that [res] is [Ok _], and returns unit.
       Otherwise, it raises some exception.
-      @raise Test_error if [res = Error _]
-      @raise Test_error if [res = Failed _] *)
+      @raise Test_fail  if [res = Failed _]
+      @raise Test_error if [res = Error _] *)
 
   type res =
     | Success
@@ -1692,17 +1848,18 @@ module Test : sig
   *)
 
   val check_cell_exn :
-    ?long:bool -> ?call:'a callback -> ?step:'a step ->
+    ?long:bool -> ?call:'a callback ->
+    ?step:'a step -> ?handler:'a handler ->
     ?rand:Random.State.t -> 'a cell -> unit
   (** Same as {!check_cell} but calls  {!check_result} on the result.
-      @raise Test_error if [res = Error _]
-      @raise Test_error if [res = Failed _] *)
+      @raise Test_fail  if [res = Failed _]
+      @raise Test_error if [res = Error _] *)
 
   val check_exn : ?long:bool -> ?rand:Random.State.t -> t -> unit
   (** Checks the property against some test cases, and calls {!check_result},
       which might raise an exception in case of failure.
-      @raise Test_error if [res = Error _]
-      @raise Test_error if [res = Failed _] *)
+      @raise Test_fail  if [res = Failed _]
+      @raise Test_error if [res = Error _] *)
 end
 
 (** {2 Sub-tests} *)
